class Vote: # Represents a single candidate paired with a preference rank
    def __init__(self, candidate, rank):
        self.candidate = candidate # A string with the candidate's name (not a candidate class)
        self.rank = rank           # The rank of the candidate for this vote
        self.counted = False       # Tells whether this vote has been counted yet

class Ballot: # One person's full ballot with all their ranked candidates
    def __init__(self, votes, id):
        self.votes = {vote.rank: vote for vote in votes}  # A list of "vote" objects
        self.id = id

class Candidate: # One candidate
    def __init__(self, name):
        self.name = name # A string with the name of the candidate
        self.total_votes = 0 # The candidate's running total of votes
        self.has_won = False # Boolean determining whether the candidate has won yet
        self.is_eligible = True # Boolean determining whether the candidate is still eligible. Set to false if candidate wins or is eliminated from the bottom of the running.

class Poll: # A full poll whose winners need to be determined
    def __init__(self, poll_name, candidates, id, num_winners):
        self.name = poll_name
        self.candidates = {candidate.name: candidate for candidate in candidates} # Pass in a list of "Candidate" objects. Makes a dictionary, key is string of candidate names, value is corresponding candidate object.
        self.ballots = []  # A list of "Ballot" objects
        self.id = id            # The poll id
        self.num_winners = num_winners # The number of winners for the given poll

    def print_all_candidate_info(self):
        for c in self.candidates.values():
            print(c.name, c.total_votes, c.has_won, c.is_eligible)

    def submit_ballot(self, ballot):
        self.ballots.append(ballot)



from math import floor
from poll import Vote, Ballot, Candidate, Poll

# Takes a Poll object, and returns a dictionary keyed with placings with values of lists of winning candidates
def run_STV_poll(poll): # Pass in a Poll object
    threshold = floor(len(poll.ballots)/(poll.num_winners + 1) + 1) # Droop quota

    """
    winners: A dictionary of all winners, to be returned.
    Key is an integer, the place.
    Value is a list (in case of ties) of strings, the winning candidates
    """
    winners = {}
    total_winners = 0 # Definite winners that have been found at a given time
    remaining_eligible_candidates = len(poll.candidates)

    for b in poll.ballots: # Loop through all the ballots and tally up the first choices
        if 1 in b.votes:
            if b.votes[1].candidate in poll.candidates:
                poll.candidates[b.votes[1].candidate].total_votes += 1
            b.votes[1].counted = True

    while total_winners < poll.num_winners:
        if remaining_eligible_candidates <= poll.num_winners - total_winners: # This ends the poll.
            while len([candidate for candidate in poll.candidates.values() if candidate.is_eligible]) > 0:
                highest_num_of_votes = max([candidate for candidate in poll.candidates.values() if candidate.is_eligible], key=lambda x:x.total_votes).total_votes
                top_eligible_candidates = [candidate for candidate in poll.candidates.values() if candidate.is_eligible and candidate.total_votes == highest_num_of_votes]
                if len(top_eligible_candidates) > 1:
                    top_eligible_candidates = break_winner_tie(top_eligible_candidates, poll)
                for c in top_eligible_candidates:
                    c.has_won = True
                    c.is_eligible = False
                    remaining_eligible_candidates -= 1
                    if total_winners + 1 not in winners:
                        winners[total_winners + 1] = []
                    winners[total_winners + 1].append(c.name)
                for c in top_eligible_candidates:
                    total_winners += 1
            #YOU ARE HERE
            return winners
        highest_num_of_votes = max([candidate for candidate in poll.candidates.values() if candidate.is_eligible], key=lambda x:x.total_votes).total_votes
        top_eligible_candidates = [candidate for candidate in poll.candidates.values() if candidate.is_eligible and candidate.total_votes == highest_num_of_votes]
        if len(top_eligible_candidates) > 1:
            top_eligible_candidates = break_winner_tie(top_eligible_candidates, poll)
        if highest_num_of_votes >= threshold: # Takes care of when there's a winner in this "voting round"
            for c in top_eligible_candidates:
                c.has_won = True
                c.is_eligible = False
                remaining_eligible_candidates -= 1
                if total_winners + 1 not in winners:
                    winners[total_winners + 1] = []
                winners[total_winners + 1].append(c.name)
            for c in top_eligible_candidates:
                total_winners += 1
                if c.total_votes > threshold:
                    redistribute_votes(c, True, poll, threshold)
        else: # Takes care of when the round has no winner, so the bottom candidate(s) are eliminated
            lowest_num_of_votes = min([candidate for candidate in poll.candidates.values() if candidate.is_eligible], key=lambda x:x.total_votes).total_votes
            bot_eligible_candidates = [candidate for candidate in poll.candidates.values() if candidate.is_eligible and candidate.total_votes == lowest_num_of_votes]
            if len(bot_eligible_candidates) > 1:
                bot_eligible_candidates = break_loser_tie(bot_eligible_candidates, poll)
            if len(bot_eligible_candidates) == remaining_eligible_candidates: # This means the poll is complete. There will be extra winners due to ties (extremely unlikely).
                for c in bot_eligible_candidates:
                    c.has_won = True
                    c.is_eligible = False
                    if total_winners + 1 not in winners:
                        winners[total_winners + 1] = []
                    winners[total_winners + 1].append(c.name)
                    return winners
            for c in bot_eligible_candidates:
                c.is_eligible = False
                remaining_eligible_candidates -= 1
            for c in bot_eligible_candidates:
                redistribute_votes(c, False, poll, threshold)
        poll.print_all_candidate_info()
    return winners

# Redistributs the votes of a candidate who is no longer eligible. Redistributes all of a bottom candidate's votes. Redistributes surplus of a winning candidate's votes, fractionally.
def redistribute_votes(cand, is_for_winner, poll, threshold):
    for b in poll.ballots:
        prev_vote_rank = max({vote.rank: vote for vote in b.votes.values() if vote.counted == True}, key=lambda x:x)
        cur_vote_rank = prev_vote_rank + 1
        if b.votes[prev_vote_rank].candidate == cand.name:
            done = False
            while done == False and cur_vote_rank in b.votes:
                    cur_vote = b.votes[cur_vote_rank]
                    if poll.candidates[cur_vote.candidate].is_eligible:
                        increment = 1;
                        if is_for_winner:
                            increment = (cand.total_votes - threshold) / cand.total_votes
                        poll.candidates[cur_vote.candidate].total_votes += increment
                        done = True
                    cur_vote.counted = True
                    cur_vote_rank += 1
    return

# Function that breaks the ties of a number of winners
def break_winner_tie(tied_cands, poll):
    n = 1           # Which rank of tied votes the candidates are on
    while len(tied_cands) > 1:
        most_votes = 0 # Running highest votes for this layer (rank)
        winning_cands = []
        for c in tied_cands: # For each candidate
            cur_cand_votes = 0
            # Tally up all the nth place votes for the current candidate across all ballots
            for b in poll.ballots: # For each ballot
                if n in b.votes and b.votes[n].candidate == c.name: # If that vote layer exists and the vote is for the current candidate
                    cur_cand_votes += 1
            # If new winner, set most_votes
            if cur_cand_votes > most_votes:
                most_votes = cur_cand_votes
                winning_cands = [c]
            # If tied with rest of candidates, just add current one to the list for next tiebreaker
            elif cur_cand_votes == most_votes:
                winning_cands.append(c)
            # Set up for next loop.
            tied_cands = winning_cands
            if n > len(poll.candidates): # In case there's a perfect tie and all votes have been taken into account
                return tied_cands
            n += 1
    return tied_cands

# Function that breaks the ties of a number of losers
def break_loser_tie(tied_cands, poll):
    n = 1           # Which rank of tied votes the candidates are on
    while len(tied_cands) > 1:
        fewest_votes = len(poll.ballots) + 1 # Running lowest votes for this layer (rank)
        losing_cands = []
        for c in tied_cands: # For each candidate
            cur_cand_votes = 0
            # Tally up all the nth place votes for the current candidate across all ballots
            for b in poll.ballots: # For each ballot
                if n in b.votes and b.votes[n].candidate == c.name: # If that vote layer exists and the vote is for the current candidate
                    cur_cand_votes += 1
            # If new loser, set fewest_votes
            if cur_cand_votes < fewest_votes:
                fewest_votes = cur_cand_votes
                losing_cands = [c]
            # If tied with rest of candidates, just add current one to the list for next tiebreaker
            elif cur_cand_votes == fewest_votes:
                losing_cands.append(c)
            # Set up for next loop.
            tied_cands = losing_cands
            if n > len(poll.candidates): # In case there's a perfect tie and all votes have been taken into account
                return tied_cands
            n += 1
    return tied_cands



from pref_vote.poll import Vote, Ballot, Poll, Candidate
from pref_vote.STV import run_STV_poll

def test_Vote():
    uut = Vote("Sean", 2)
    assert(uut.candidate == "Sean")
    assert(uut.rank == 2)
    assert (uut.counted == False)

def test_Ballot():
    test_Votes = [Vote("Sean", 1), Vote("Neil", 2)]
    uut = Ballot(test_Votes, 1234)
    assert(uut.votes[1] == test_Votes[0])
    assert(uut.votes[2] == test_Votes[1])

def test_Candidate():
    uut = Candidate("Neil")
    assert(uut.name == "Neil")
    assert(uut.total_votes == 0)
    assert(uut.has_won == False)
    assert(uut.is_eligible == True)

def test_Poll():
    test_Votes_1 = [Vote("Sean", 1), Vote("Neil", 2)]
    test_Votes_2 = [Vote("John", 2), Vote("Cob", 1)]
    test_Ballots = [Ballot(test_Votes_1, 1234), Ballot(test_Votes_2, 5678)]
    test_Candidates = [Candidate("Sean"), Candidate("Neil"), Candidate("John"), Candidate("Cob")]
    uut = Poll('foo', test_Candidates, 9012, 1)
    for ballot in test_Ballots:
        uut.submit_ballot(ballot)

    assert(uut.candidates["Sean"] == test_Candidates[0])
    assert(uut.candidates["Neil"] == test_Candidates[1])
    assert(uut.candidates["John"] == test_Candidates[2])
    assert(uut.candidates["Cob"] == test_Candidates[3])

    assert(uut.ballots == test_Ballots)
    assert(uut.id == 9012)
    assert(uut.num_winners == 1)

def test_simple_poll_1():
    test_Votes_1 = [Vote("A", 1), Vote("B", 2), Vote("C", 3)]
    test_Ballots = [Ballot(test_Votes_1, 1)]
    test_Candidates = [Candidate("A"), Candidate("B"), Candidate("C")]
    test_Poll = Poll('foo', test_Candidates, 9012, 1)
    for ballot in test_Ballots:
        test_Poll.submit_ballot(ballot)

    assert(run_STV_poll(test_Poll) == {1:["A"]})

def test_simple_poll_2():
    test_Votes_1 = [Vote("A", 1), Vote("B", 2), Vote("C", 3)]
    test_Ballots = [Ballot(test_Votes_1, 1)]
    test_Candidates = [Candidate("A"), Candidate("B"), Candidate("C")]
    test_Poll = Poll('foo', test_Candidates, 9012, 2)
    for ballot in test_Ballots:
        test_Poll.submit_ballot(ballot)

    assert(run_STV_poll(test_Poll) == {1:["A"], 2:["B"]})

def test_simple_poll_3():
    test_Votes_1 = [Vote("A", 1), Vote("B", 2), Vote("C", 3)]
    test_Ballots = [Ballot(test_Votes_1, 1)]
    test_Candidates = [Candidate("A"), Candidate("B"), Candidate("C")]
    test_Poll = Poll('foo', test_Candidates, 9012, 3)
    for ballot in test_Ballots:
        test_Poll.submit_ballot(ballot)

    assert(run_STV_poll(test_Poll) == {1:["A"], 2:["B"], 3:["C"]})

def test_more_winners_than_candidates():
    test_Votes_1 = [Vote("A", 1), Vote("B", 2), Vote("C", 3)]
    test_Ballots = [Ballot(test_Votes_1, 1)]
    test_Candidates = [Candidate("A"), Candidate("B"), Candidate("C")]
    test_Poll = Poll('foo', test_Candidates, 9012, 4)
    for ballot in test_Ballots:
        test_Poll.submit_ballot(ballot)

    assert(run_STV_poll(test_Poll) == {1:["A"], 2:["B"], 3:["C"]})

def test_equal_number_winners_and_candidates():
    pass

def test_perfect_loser_tie():
    pass

def test_imperfect_loser_tie():
    pass

def test_perfect_winner_tie():
    pass

def test_imperfect_winner_tie():
    pass

def test_more_winners_than_allocated():
    pass




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Setup file for pref_vote.

    This file was generated with PyScaffold 2.5.8, a tool that easily
    puts up a scaffold for your new Python project. Learn more under:
    http://pyscaffold.readthedocs.org/
"""

from setuptools import setup


setup(name='pref_vote',
      description='Preferential voting tool',
      version='0.0.4',
      url='https://github.com/jmhossler/pref_vote',
      author='John Hossler',
      author_email='john.m.hossler@gmail.com',
)
